module ScopeChecker (scopeCheck) where

import Abstract (Name,mkExtRef,Co,Pos,MVar)
import qualified Abstract as A
import qualified Concrete as C

import TraceError

import Control.Monad.Identity
import Control.Monad.Reader
import Control.Monad.State
import Control.Monad.Error

import Data.List

import Debug.Trace

import Util

--scope checker
-- check that all identifiers are in scope and global identifiers are only used once
-- replaces Ident with Con, Def , Let or Var  
-- replaces IdentP with ConP or VarP in patterns
-- replaces Unknown by a new Meta-Variable
-- check pattern length is equal in each clause
-- group mutual declarations

--------------------


--local environment
type Context = [Name] 


emptyCtx :: Context
emptyCtx = []

lookupCtx :: Name -> Context -> Bool
lookupCtx n [] = False
lookupCtx n (x:xs) = if (x == n) then True else lookupCtx n xs 

addCtx :: Name -> Context -> Context
addCtx n ctx = ctx ++ [n]

addCtxs :: [Name] -> Context -> Context
addCtxs nl ctx = ctx ++ nl


-- kind of identifier
data IKind = DataK 
           | ConK Co 
           | FunK Bool  -- False = inside body, True = outside body 
           | LetK

-- Scope check signature
type Sig = [(Name,IKind)] 

emptySig :: Sig
emptySig= []

lookupSig :: Name -> Sig -> Maybe IKind
lookupSig n [] = Nothing
lookupSig n ((x,k):xs) = if (x == n) then Just k else lookupSig n xs

-- Scope check monad
-- reader monad for local environment for variables (used in expresssions and patterns)
-- state monad for global signature

data SCState = SCState { signature :: Sig, nextMeta :: MVar }

initSt = SCState emptySig 0

type ScopeCheck a = ReaderT Context (StateT SCState (ErrorT TraceError Identity)) a

nextMVar :: (MVar -> ScopeCheck a) -> ScopeCheck a
nextMVar f = do
  st <- get
  put $ st { nextMeta = nextMeta st + 1 }
  f (nextMeta st)

addSig :: Name -> IKind -> ScopeCheck ()
addSig n k = do
  st <- get
  put $ st { signature = (n,k) : signature st }  -- ++ [(n,k)] }


runScopeCheck :: Context -> SCState -> ScopeCheck a -> Either TraceError (a,SCState)
runScopeCheck ctx st sc = runIdentity (runErrorT (runStateT (runReaderT sc ctx) st))  

scopeCheck :: [C.Declaration] -> Either TraceError ([A.Declaration],SCState)
scopeCheck dl = runScopeCheck emptyCtx initSt (scopeCheckDecls dl)

---------

scopeCheckDecls :: [C.Declaration] -> ScopeCheck [A.Declaration]
scopeCheckDecls = mapM scopeCheckDeclaration 
                       

scopeCheckDeclaration :: C.Declaration -> ScopeCheck A.Declaration
scopeCheckDeclaration d@(C.DataDecl _ _ _ _ _ _) = scopeCheckDataDecl d 
scopeCheckDeclaration d@(C.FunDecl co _ _) = scopeCheckFunDecls co [d]
                                               
scopeCheckDeclaration (C.LetDecl b ts e) = do e' <- scopeCheckExpr e
                                              ts' <- scopeCheckTypeSig ts
                                              addTypeSig LetK ts 
                                              return $ A.LetDecl b ts' e'

-- we support
-- - mutual (co)funs
-- - mutual (co)data

scopeCheckDeclaration (C.MutualDecl []) = throwErrorMsg "mutual combination not supported"
scopeCheckDeclaration (C.MutualDecl l@(C.FunDecl  co _ _:xl)) = scopeCheckFunDecls co l 
scopeCheckDeclaration (C.MutualDecl _) = throwErrorMsg "mutual combination not supported"
                                
scopeCheckDataDecl :: C.Declaration -> ScopeCheck A.Declaration
scopeCheckDataDecl decl@(C.DataDecl n sz co tel t cs) = 
    (
    do
      let pos = posTelescope tel
      let tt = C.teleToType tel t
      (A.TypeSig _ tt') <- scopeCheckTypeSig (C.TypeSig n tt)
      let (tel',t') = A.typeToTele' (length tel) tt'                
      addDecl decl
      let names = collectTelescopeNames tel
      cs' <- local (addCtxs names) (mapM (scopeCheckConstructor co) cs) 
      -- traceM ("constructors: " ++ show cs')
--      when (t' == A.Sort A.Set && length cs' == 1) $ do
      when (length cs' == 1) $ do
-- Add destructor names
        let fields = A.classifyFields co n (A.typePart (head cs'))
        -- TODO Check for recursive occurrence!
        -- when (A.etaExpandable fields) $
        let destrNames =  A.destructorNames fields
        --when (not (null (destrNames))) $ 
        -- traceM ("fields: " ++ show fields)
        -- traceM ("destructors: " ++ show destrNames) 
        mapM_ (addName (FunK True)) $ destrNames -- destructors are also upped
 {- 
        let (ctel,_) = A.typeToTele (A.typePart (head cs'))
        let destrNames = map (\(_,x,_) -> x) ctel
        when (all (/= "") destrNames) $
          mapM_ (addName (FunK True)) destrNames -- destructors are also upped
-}
      return $ A.DataDecl n sz co pos tel' t' cs'
    ) `throwTrace` n


checkFunMutual co [] = return ()
checkFunMutual co (C.FunDecl co' _ _:xl) | co == co' = checkFunMutual co xl
checkFunMutual _ _ = throwErrorMsg "mutual combination not supported"

posTelescope :: C.Telescope -> [A.Pos]
posTelescope = map (A.polarity . C.boundDec)

---
scopeCheckFunDecls co l = 
    do checkFunMutual co l
       tsl' <- mapM scopeCheckFunSig l
       mapM (addFunDecl False) l
       cll' <- mapM scopeCheckFunClause l
       mapM (addFunDecl True) l
       return $ A.FunDecl co (zip tsl' cll') 

scopeCheckFunSig :: C.Declaration -> ScopeCheck A.TypeSig 
scopeCheckFunSig (C.FunDecl _ ts _ ) = scopeCheckTypeSig ts

scopeCheckFunClause :: C.Declaration -> ScopeCheck [A.Clause] 
scopeCheckFunClause (C.FunDecl _ (C.TypeSig n _) cl) =
    (
    do cl' <- mapM (scopeCheckClause (Just n)) cl
       let b = checkPatternLength cl
       case b of
          True -> return $ cl' 
          False -> throwErrorMsg $ " pattern length differs"
    ) `throwTrace` n
scopeCheckTypeSig :: C.TypeSig -> ScopeCheck A.TypeSig 
scopeCheckTypeSig a@(C.TypeSig n t) = 
    (
    do sig <- gets signature
       case (lookupSig n sig) of
         Just _ -> errorAlreadyInSignature a n
         Nothing -> do t' <- scopeCheckExpr t 
                       return $ A.TypeSig n t'         
    ) `throwTrace` n

addDecl :: C.Declaration -> ScopeCheck ()
addDecl (C.DataDecl n _ _ _ _ _) = addName DataK n

addFunDecl :: Bool -> C.Declaration -> ScopeCheck () 
addFunDecl b (C.FunDecl _ ts _) = addTypeSig (FunK b) ts

addTypeSig :: IKind -> C.TypeSig -> ScopeCheck ()
addTypeSig kind (C.TypeSig n _) = addName kind n

addName :: IKind -> Name -> ScopeCheck ()
addName kind n =  addSig n kind

collectTelescopeNames :: C.Telescope -> [Name]
collectTelescopeNames = map C.boundName

scopeCheckConstructor :: Co -> C.Constructor -> ScopeCheck A.Constructor
scopeCheckConstructor co a@(C.TypeSig n t) = 
    do sig <- gets signature
       case (lookupSig n sig) of
         Just _ -> errorAlreadyInSignature a n
         Nothing -> do t' <- scopeCheckExpr t 
                       addSig n (ConK co)
                       return $ A.TypeSig n t'

                     
scopeCheckExpr :: C.Expr -> ScopeCheck A.Expr
scopeCheckExpr e = 
    case e of 
      -- replace underscore by next meta-variable
      C.Unknown -> nextMVar (return . A.Meta)
{-
      C.Set -> return $ A.Sort $ A.Set $ A.Zero
      C.Type e -> scopeCheckExpr e >>= return . A.Sort . A.Set
-}
      C.Set e -> scopeCheckExpr e >>= return . A.Sort . A.Set
      C.CoSet e -> scopeCheckExpr e >>= return . A.Sort . A.CoSet
      C.Size -> return $ A.Sort A.Size
      C.Succ e1 -> do e1' <- scopeCheckExpr e1
                      return $ A.Succ e1'
      C.Zero -> return A.Zero
      C.Infty -> return A.Infty
      C.Sing e1 et -> do e1' <- scopeCheckExpr e1
                         et' <- scopeCheckExpr et
                         return $ A.Sing e1' et'
      C.App C.Max el -> do el' <- mapM scopeCheckExpr el
                           when (length el' <  2) $ throwErrorMsg "max expects at least 2 arguments"
                           return $ A.Max el'
      C.App e1 el -> do e1' <- scopeCheckExpr e1
                        el' <- mapM scopeCheckExpr el
                        return $ foldl A.App e1' el'
      C.Case e cl -> do 
        e'  <- scopeCheckExpr e
        cl' <- mapM (scopeCheckClause Nothing) cl
        return $ A.Case e' cl'

      C.Pi (C.TBind dec "" t) e1 -> do 
        t' <- scopeCheckExpr t -- non-dep. function type
        e1' <- scopeCheckExpr e1
        return $ A.Pi (A.TBind dec "" t') e1'

      -- interesting cases
      C.Pi (C.TBind dec n t) e1 -> do 
        ctx <- ask
        sig <- gets signature
        case (lookupSig n sig) of
          Just _ ->  errorAlreadyInSignature e n 
          Nothing -> case (lookupCtx n ctx) of 
                      True  -> errorAlreadyInContext e n
                      False -> do t' <- scopeCheckExpr t
                                  e1' <- local (addCtx n) (scopeCheckExpr e1)
                                  return $ A.Pi (A.TBind dec n t') e1'

      C.Lam n e1 -> do 
        ctx <- ask
        sig <- gets signature
        case (lookupSig n sig) of
          Just _ ->  errorAlreadyInSignature e n
          Nothing -> case (lookupCtx n ctx) of 
                       True  -> errorAlreadyInContext e n
                       False -> do e1' <- local (addCtx n) (scopeCheckExpr e1)
                                   return $ A.Lam A.defaultDec n e1'

      C.LLet (C.TBind dec n t1) e1 e2 ->  do
        ctx <- ask
        sig <- gets signature
        case (lookupSig n sig) of
          Just _ ->  errorAlreadyInSignature e n 
          Nothing -> case (lookupCtx n ctx) of 
                       True  -> errorAlreadyInContext e n
                       False -> do t1' <- scopeCheckExpr t1
                                   e1' <- scopeCheckExpr e1
                                   e2' <- local (addCtx n) (scopeCheckExpr e2)
                                   return $ A.LLet (A.TBind dec n t1') e1' e2'
      C.Ident n -> do ctx <- ask
                      sig <- gets signature
                      case (lookupSig n sig) of
                        Just k -> case k of
                                    (ConK co) -> return $ A.con co n
                                    LetK -> return $ A.letdef n
-- references to recursive functions are coded differently outside the
-- mutual block
                                    FunK True -> return $ A.mkExtRef n
                                    FunK False -> return $ A.fun n
                                    DataK -> return $ A.dat n  
                        Nothing -> case (lookupCtx n ctx) of
                                     True -> return $ A.Var n
                                     False -> errorIdentifierUndefined n 


scopeCheckClause :: Maybe Name -> C.Clause -> ScopeCheck A.Clause
scopeCheckClause mname' (C.Clause mname pl mrhs) = 
    do when (mname /= mname') $ errorClauseIdentifier mname mname'
       pl' <- mapM scopeCheckPattern pl
       names <- collectVarPNames pl'
       pl'' <- local (addCtxs names) (zipWithM scopeCheckDotPattern pl pl') 
       case mrhs of
         Nothing -> return $ A.Clause pl'' Nothing
         Just rhs -> do
           rhs' <- local (addCtxs names) (scopeCheckExpr rhs)
           return $ A.Clause pl'' (Just rhs')

checkPatternLength :: [C.Clause] -> Bool
checkPatternLength [] = True
checkPatternLength (C.Clause _ pl _:cl) = cpl (length pl) cl
 where
   cpl k [] = True
   cpl k (C.Clause _ pl _ : cl) = if (length pl == k) then (cpl k cl) else False

-- collect all variable names, checks if pattern is linear
collectVarPNames :: [A.Pattern] -> ScopeCheck [Name]
collectVarPNames pl = cvp pl [] where
    cvp [] acc = return acc
    cvp (p:pl) acc = case p of
                       A.AbsurdP -> cvp pl acc
                       A.SuccP p2 -> cvp (p2:pl) acc
                       A.ConP _ n pl2 -> cvp (pl2++pl) acc
                       A.VarP "" -> cvp pl acc -- will be dot pattern
                       A.VarP n ->  if (elem n acc) then 
                                      errorPatternNotLinear n
                                    else 
                                        cvp pl (n:acc)
                       A.SizeP m n ->  
                                    if (elem n acc) then 
                                      errorPatternNotLinear n
                                    else 
                                        cvp pl (n:acc)
          

scopeCheckPattern :: C.Pattern -> ScopeCheck A.Pattern
scopeCheckPattern p = 
    case p of
      C.IdentP n -> do sig <- gets signature
                       ctx <- ask
                       case (lookupSig n sig) of
                         (Just (ConK co)) -> return $ A.ConP co n [] -- a nullary constructor
                         (Just _) -> errorPatternNotConstructor n
                         Nothing -> return $ A.VarP n
      C.SizeP m n-> do -- cxt <- ask
                       -- when (not $ lookupCtx m cxt) $ -- check this later! (dotP)
                       --   errorIdentifierUndefined m
                       sig <- gets signature
                       case (lookupSig n sig) of
                         (Just _) -> errorPatternNotConstructor n
                         Nothing -> return $ A.SizeP m n
      C.SuccP p2 -> do p2' <- scopeCheckPattern p2
                       return $ A.SuccP p2'
      C.ConP n pl -> do sig <- gets signature
                        case (lookupSig n sig) of
                          Just (ConK co) -> do
                              pl' <- mapM scopeCheckPattern pl 
                              return $ A.ConP co n pl'
                          Just _ -> errorPatternNotConstructor n 
                          Nothing -> errorPatternNotConstructor n 
      C.DotP p -> return $ A.VarP "" -- dot patterns checked later
      C.AbsurdP -> return $ A.AbsurdP
    
scopeCheckDotPattern :: C.Pattern -> A.Pattern -> ScopeCheck A.Pattern
scopeCheckDotPattern p1 p2 = 
    case p1 of 
      C.SizeP m n -> do cxt <- ask
                        when (not $ lookupCtx m cxt) $
                          errorIdentifierUndefined m
                        return p2
      C.DotP e -> do e' <- scopeCheckExpr e
                     return $ A.DotP e'
      C.SuccP p1' -> do let (A.SuccP p2') = p2
                        p2'' <- scopeCheckDotPattern p1' p2'
                        return $ A.SuccP p2''
      (C.ConP _ pl) -> 
            do let (A.ConP co n pl') = p2
               pl'' <- zipWithM scopeCheckDotPattern pl pl'
               return $ A.ConP co n pl''
      _ -> return p2


errorAlreadyInSignature s n = throwErrorMsg $ show s  ++ ": Identifier " ++ n ++ " already in signature"

errorAlreadyInContext s n = throwErrorMsg $ show s ++ ": Identifier " ++ n ++ " already in context"

errorPatternNotConstructor n = throwErrorMsg $ "pattern " ++ n ++ " not a constructor"

errorIdentifierUndefined n = throwErrorMsg $ "Identifier " ++ n ++ " undefined"

errorPatternNotLinear n = throwErrorMsg $ "pattern not linear: " ++ n 

errorClauseIdentifier (Just n) (Just n') = throwErrorMsg $ "Expected identifier " ++ show n' ++ " as clause head, found " ++ show n