{-# LANGUAGE TypeSynonymInstances, PatternGuards #-}
 
module TCM where

import Control.Monad.Identity
import Control.Monad.State
import Control.Monad.Error
import Control.Monad.Reader

import Control.Applicative
import Data.Foldable (Foldable)
import qualified Data.Foldable as Foldable
import Data.Traversable (Traversable) 
import qualified Data.Traversable as Traversable
import Data.Monoid

import Data.Map (Map)
import qualified Data.Map as Map

import Debug.Trace

import Abstract
import Value
import {-# SOURCE #-} Eval -- (up,whnf')

-- import CallStack 
import TraceError

import TreeShapedOrder (TSO)
import qualified TreeShapedOrder as TSO

import Util

import Warshall

{-
traceRew msg a = a -- trace msg a 
traceRewM msg = return () -- traceM msg
-}
traceRew msg a = trace msg a 
traceRewM msg = traceM msg

data OneOrTwo a = One a | Two a a deriving (Eq, Ord)

instance Show a => Show (OneOrTwo a) where
  show (One a)   = show a
  show (Two a b) = show a ++ "||" ++ show b

instance Functor OneOrTwo where
  fmap f (One a)   = One (f a)
  fmap f (Two a b) = Two (f a) (f b)

instance Foldable OneOrTwo where
  foldMap f (One a) = f a
  foldMap f (Two a b) = f a `mappend` f b

-- traverse :: Applicative f => (a -> f b) -> t a -> f (t b)	
instance Traversable OneOrTwo where
  traverse f (One a) = One <$> f a
  traverse f (Two a b) = Two <$> f a <*> f b

type Env2 = Environ (OneOrTwo Val)

-- reader monad for local environment (not used right now)

data TCContext = TCContext 
  { context  :: SemCxt
  , renaming :: Ren
  , environ  :: Env2
  , rewrites :: Rewrites
  , sizeRels :: TSO Int   -- relations of universal (rigid) size variables
                          -- collected from size patterns (x > y)
  , checkingConType :: Bool  -- different PTS rules for constructor types (parametric function space!)
  }

instance Show TCContext where
    show ce = show (environ ce) ++ "; " ++ show (context ce)

emptyContext = TCContext 
  { context  = cxtEmpty
  , renaming = Map.empty 
  , environ  = emptyEnv 
  , rewrites = emptyRewrites
  , sizeRels = TSO.empty
  , checkingConType = False
  }

-- state monad for global signature

data TCState = TCState {
    signature   :: Signature,
    constraints :: Constraints
  }

-- type TypeCheck = StateT TCState (ReaderT TCContext (CallStackT String IO)) 
type TypeCheck = StateT TCState (ReaderT TCContext (ErrorT TraceError IO)) 


-- rewriting rules -----------------------------------------------

data Rewrite  = Rewrite { lhs :: Val,  rhs :: Val }
type Rewrites = [Rewrite]

emptyRewrites = []

instance Show Rewrite where
  show rr = show (lhs rr) ++ " --> " ++ show (rhs rr)

-- renaming ------------------------------------------------------

type Ren = Map Name Int


{- context -------------------------------------------------------

A context maps generic values to their type value.

During type checking, named variables are mapped to
generic values via a renaming.  Thus, looking up the type of a
name involves first looking up the generic value, and then its type.

-}

data Domain = Domain { typ :: TVal, decor :: Dec } 
-- TODO: data Domain = Domain { typ :: TVal, kind :: Sort, decor :: Dec } 

instance Show Domain where
  show item = (if erased (decor item) then brackets else id) (show (typ item))

-- During heterogeneous equality, a variable might have
-- two different types, one on the left and one on the right.
-- We implement this as Two tl tr.

type CxtEntry = (OneOrTwo TVal, Dec)

data SemCxt = SemCxt { len :: Int, cxt :: Map Int (OneOrTwo TVal), decs :: Map Int Dec }
{- invariant: length (cxt delta) = length (decs delta) = len -}

instance Show SemCxt where
  show delta = show $ 
    zipWith (\ entry dec -> fmap ((flip Domain) dec) entry) 
      (Map.elems (cxt delta))
      (Map.elems (decs delta))

cxtEmpty = SemCxt { len = 0, cxt = Map.empty, decs = Map.empty }

-- push a new type declaration on context
cxtPush' :: CxtEntry -> SemCxt -> SemCxt
cxtPush' (tv12, dec) delta = 
  delta { len = k + 1
        , cxt  = Map.insert k tv12 (cxt delta)
        , decs = Map.insert k dec (decs delta) }
  where k = len delta

cxtPush :: Dec -> TVal -> SemCxt -> (Int, SemCxt)
cxtPush dec v delta = (len delta, cxtPush' (One v, dec) delta)

cxtPushEntry :: CxtEntry -> SemCxt -> (Int, SemCxt)
cxtPushEntry ce delta = (len delta, cxtPush' ce delta)

-- push a variable with a left and a right type
cxtPush2 :: Dec -> TVal -> TVal -> SemCxt -> (Int, SemCxt)
cxtPush2 dec tvl tvr delta = 
  (len delta, cxtPush' (Two tvl tvr, dec) delta)

cxtPushGen ::  SemCxt -> (Int, SemCxt)
cxtPushGen delta = cxtPush bot bot delta
  where bot = error "IMPOSSIBLE: this generic is not bound to any type"

-- only defined for single bindings
cxtSetType :: Int -> Dec -> TVal -> SemCxt -> SemCxt
cxtSetType k dec tv delta = 
  delta { cxt  = Map.insert k (One tv) (cxt delta)
        , decs = Map.insert k dec (decs delta) }

cxtLookupGen :: Monad m => SemCxt -> Int -> m Domain
cxtLookupGen delta k = do
  One tv <- lookupM k (cxt delta)
  dec    <- lookupM k (decs delta)
  return $ Domain { typ = tv, decor = dec }

cxtLookupName :: Monad m => SemCxt -> Ren -> Name -> m Domain
cxtLookupName delta ren x = do
  i <- lookupM x ren
  cxtLookupGen delta i

-- apply decoration, possibly resurrecting (see Pfenning, LICS 2001)
-- and changing polarities (see Abel, MSCS 2008)
cxtApplyDec :: Dec -> SemCxt -> SemCxt
cxtApplyDec dec delta =  delta { decs = Map.map (invCompDec dec) (decs delta) }

-- clear all "erased" flags (see Pfenning, LICS 2001)
cxtResurrect :: SemCxt -> SemCxt
cxtResurrect delta = delta { decs = Map.map resurrectDec (decs delta) }


-- manipulating the context ------------------------------------------

class Monad m => MonadCxt m where
--  bind     :: Name -> Domain -> Val -> m a -> m a
--  new performs eta-expansion "up" of new gen 
  newVar     :: Name -> CxtEntry -> (Int -> OneOrTwo Val -> m a) -> m a 
  newWithGen :: Name -> Domain -> (Int -> Val -> m a) -> m a 
  newWithGen x (Domain tv dec) k = newVar x (One tv, dec) 
    (\ i (One v) -> k i v)
  new2WithGen:: Name -> (TVal, TVal, Dec) -> (Int -> (Val, Val) -> m a) -> m a   
  new2WithGen x (tvl, tvr, dec) k = newVar x (Two tvl tvr, dec) 
    (\ i (Two vl vr) -> k i (vl, vr))
  new        :: Name -> Domain -> (Val -> m a) -> m a
  new x d cont = newWithGen x d (\ _ -> cont)   
  new2       :: Name -> (TVal, TVal, Dec) -> ((Val, Val) -> m a) -> m a
  new2 x d cont = new2WithGen x d (\ _ -> cont)   
  new'       :: Name -> Domain -> m a -> m a
  new' x d cont = new x d (\ _ -> cont) 
  addName    :: Name -> (Val -> m a) -> m a   
--  addName x = new x dontCare
  setType    :: Int -> Domain -> m a -> m a
  lookupGen  :: Int -> m Domain
  lookupGenType :: Int -> m TVal
  lookupGenType i = lookupGen i >>= return . typ
  lookupName :: Name -> m Domain
  getLen     :: m Int       -- return length of the context
  getEnv     :: m Env       -- return current environment
  getRen     :: m Ren       -- return current renaming
  applyDec   :: Dec -> m a -> m a  -- resurrect/adjust polarities
  resurrect  :: m a -> m a -- resurrect all erased variables in context
  addRewrite :: Rewrite -> [Val] -> ([Val] -> m a) -> m a
  addPattern :: TVal -> Pattern -> Env -> (TVal -> Val -> Env -> m a) -> m a -- step under pat
  addPatterns:: TVal -> [Pattern] -> Env -> (TVal -> [Val] -> Env -> m a) -> m a
  addSizeRel :: Int -> Int -> m a -> m a  
  getSizeDiff :: Int -> Int -> m (Maybe Int)
  checkingCon :: Bool -> m a -> m a
  checkingDom :: m a -> m a  -- check domain A of Pi x:A.B (takes care of polarities)

dontCare = error "Internal error: tried to retrieve unassigned type of variable" 

instance MonadCxt TypeCheck where

  -- UPDATE to 2?
  addName x f = enter ("new " ++ x ++ " : _") $ do
    cxtenv <- ask
    let (k, delta) = cxtPushGen (context cxtenv)
    let v = VGen k
    let rho = update (environ cxtenv) x (One v) 
    local (\ cxt -> cxt { context = delta
                        , renaming = Map.insert x k (renaming cxtenv)
                        , environ = rho }) (f v)

  newVar x (tv12, dec) f = enter ("new " ++ x ++ " : " ++ show tv12) $ do
    cxtenv <- ask
    let (k, delta) = cxtPushEntry (tv12, dec) (context cxtenv)
    v12 <- Traversable.mapM (up (VGen k)) tv12
    let rho = update (environ cxtenv) x v12 
    local (\ cxt -> cxt { context = delta
                        , renaming = Map.insert x k (renaming cxtenv)
                        , environ = rho }) (f k v12)

  setType k (Domain tv dec) = 
    local (\ ce -> ce { context = cxtSetType k dec tv (context ce) }) 

  lookupGen k = do
    ce <- ask
    cxtLookupGen (context ce) k

  lookupName x = do
    ce <- ask
    cxtLookupName (context ce) (renaming ce) x

  getLen = do
    ce <- ask
    return $ len (context ce)

  getRen = do
    ce <- ask
    return $ renaming ce

  -- quick hack for now, just implement for One
  getEnv = do
    ce <- ask
    return $ map (\ (x, One v) -> (x, v)) $ environ ce

  applyDec dec = local (\ ce -> ce { context = cxtApplyDec dec (context ce) })

  resurrect cont = do
    ce <- ask
    local (\ _ -> ce { context = cxtResurrect (context ce) }) cont


  -- PROBABLY TOO INEFFICIENT
  addRewrite rew vs cont = traceRew ("adding rewrite " ++ show rew) $
    -- add rewriting rule
    local (\ cxt -> cxt { rewrites = rew : (rewrites cxt) }) $ do
      ce <- ask
      -- normalize all types in context
      traceRewM "normalizing types in context" 
      cx' <- mapMapM (Traversable.mapM reval) (cxt (context ce))  -- LOOP!
      -- normalize environment 
      traceRewM "normalizing environment"
      en' <- mapM (\ (x,v12) -> Traversable.mapM reval v12 >>= \ v12' -> return (x, v12')) (environ ce) 
      -- normalize given values 
      vs' <- mapM reval vs
      -- continue in updated context
      local (\ ce -> ce { context = (context ce) { cxt = cx' }
                        , environ = en' }) $ cont vs' 

  -- addPattern :: TVal -> Pattern -> (TVal -> Val -> Env -> m a) -> m a
  addPattern tv@(VPi dec x av env b) p rho cont =  
       case p of
          VarP y -> new y (Domain av dec) $ \ xv -> do
              bv <- whnf (update env x xv) b
              cont bv xv ((y,xv):rho)

          SizeP z y -> newWithGen y (Domain av dec) $ \ j xv -> do
              bv <- whnf (update env x xv) b
              VGen k <- whnf' (Var z) 
              addSizeRel j k $
                cont bv xv ((y,xv):rho)

          ConP co n pl -> do
              sig <- gets signature 
              cxt <- ask
              let vc = symbTyp (lookupSig n sig)
              addPatterns vc pl rho $ \ vc' vpl rho -> do
                pv0 <- foldM app (vCon co n) vpl
                pv  <- up pv0 av
                vb  <- whnf (update env x pv) b
                cont vb pv rho

          SuccP p2 -> do  
              cxt <- ask
              addPattern (VSort Size `arrow` VSort Size) p2 rho $ \ _ vp2 rho -> do
                let pv = succSize vp2 
                vb  <- whnf (update env x pv) b
                cont vb pv rho

          ErasedP p -> addPattern tv p rho cont

-- for dot patterns, we have to do something smart, because they might
-- contain identifiers which are not yet in scope, only after adding
-- other patterns
-- the following trivial solution only works for trivial dot patterns, i.e.,
-- such that do not use yet undeclared identifiers

          DotP e -> do
              v  <- whnf rho e
              vb <- whnf (update env x v) b
              cont vb v rho -- [(x,v)]

   
  addPatterns tv [] rho cont = cont tv [] rho
  addPatterns tv (p:ps) rho cont = 
    addPattern tv p rho $ \ tv' v env -> 
      addPatterns tv' ps env $ \ tv'' vs env' -> 
        cont tv'' (v:vs) env' -- (env' ++ env)

  addSizeRel son father k = enter ("adding size rel. " ++ show son ++ " < " ++ show father) $ do
    local (\ cxt -> cxt { sizeRels = TSO.insert son father (sizeRels cxt) }) k

  getSizeDiff son ancestor = do
    cxt <- ask
    return $ TSO.diff son ancestor (sizeRels cxt)

  checkingCon b = local (\ cxt -> cxt { checkingConType = b})

  checkingDom = local $ \ cxt -> 
    if checkingConType cxt then cxt
     else cxt { context = cxtApplyDec (Dec False Neg) (context cxt) }

addBind :: TBind -> TypeCheck a -> TypeCheck a
addBind (TBind dec x t) cont = do
  tv <- whnf' t
  new' x (Domain tv dec) cont

addBinds :: Telescope -> TypeCheck a -> TypeCheck a
addBinds tel k0 = foldr addBind k0 tel

-- signature -----------------------------------------------------
-- input to and output of the type-checker

type Signature = Map.Map Name SigDef

-- a signature entry is either  
-- * a fun/cofun, 
-- * a defined constant, 
-- * a constructor, or 
-- * a data type id with its kind
-- they share "symbTyp", the type signature of the definition
data SigDef 
  = FunSig  { isCo          :: Co 
            , symbTyp       :: TVal 
            , clauses       :: [Clause] 
            , isTypeChecked :: Bool 
            } --type , co , clauses , whether its type checked
  | LetSig  { symbTyp       :: TVal 
            , definingVal   :: Val 
--            , definingExpr  :: Expr 
            }-- type , expr 
  | ConSig  { numPars       :: Int
            , isSized       :: Sized
            , recOccs       :: [Bool] -- which of the arguments contain rec.occs.of the (co)data type?
            , symbTyp       :: TVal } -- type   
  | DataSig { numPars       :: Int
            , positivity    :: [Pos] 
            , isSized       :: Sized 
            , isCo          :: Co
            , symbTyp       :: TVal
            -- the following information is only needed for eta-expansion
            -- hence it is only provided for suitable ind.fams.
            , constructors  :: [ConstructorInfo]
            , etaExpand     :: Bool -- non-overlapping pattern inductive family
                                    -- with at least one eta-expandable constructor
            , isTuple       :: Bool -- exactly one constructor, can be recursive
                                    -- must be pattern inductive family
                                    -- qualifies for target of corecursive fun
{-
            , destructors   :: Maybe [Name] -- Nothing if not a record
            , isFamily      :: Bool 
-}
            } -- # parameters, positivity of parameters  , sized , co , type  
              deriving (Show)

emptySig = Map.empty

lookupSymb :: Name -> TypeCheck SigDef
lookupSymb n = do
  sig <- gets signature
  return $ lookupSig n sig 

lookupSig :: Name -> Signature -> SigDef
lookupSig n sig = case (Map.lookup n sig) of
                       Nothing -> error $ "Error not in signature: "  ++ n ++ " " ++ show sig
                       Just k -> k

addSig :: Name -> SigDef -> TypeCheck ()
addSig n def = do
  st <- get
  put $ st { signature = Map.insert n def $ signature st }

modifySig :: Name -> (SigDef -> SigDef) -> TypeCheck ()
modifySig n f = do
  st <- get
  put $ st { signature = Map.adjust f n $ signature st }

-- more on the type checking monad -------------------------------

initSt :: TCState
initSt = TCState emptySig emptyConstraints

resetConstraints :: TypeCheck ()
resetConstraints = do
  st <- get
  put $ st { constraints = emptyConstraints }

vGenSuccs (VGen k)  m = (k,m)
vGenSuccs (VSucc v) m = vGenSuccs v (m+1)

retret = return . return

mkConstraint :: Val -> Val -> TypeCheck (Maybe Constraint)
mkConstraint v (VMax vs) = do
  bs <- mapM (errorToBool . leqSize' v) vs
  if any id bs then return Nothing else 
   fail $ "cannot handle constraint " ++ show v ++ " <= " ++ show (VMax vs) 
mkConstraint (VMeta i n) (VMeta j m) = retret $ Arc (Flex i) (m-n) (Flex j)
mkConstraint (VMeta i n) VInfty      = retret $ Arc (Flex i) 0 (Rigid (RConst Infinite))
mkConstraint (VMeta i n) v           = retret $ Arc (Flex i) (m-n) (Rigid (RVar j))
  where (j,m) = vGenSuccs v 0
mkConstraint VInfty (VMeta i n)      = retret $ Arc (Rigid (RConst Infinite)) 0 (Flex i)
mkConstraint v (VMeta j m)           = retret $ Arc (Rigid (RVar i)) (m-n) (Flex j)
  where (i,n) = vGenSuccs v 0

-- addMeta k x  adds a metavariable which can refer to VGens < k
addMeta :: Int -> MVar -> TypeCheck ()
addMeta k i = do
  st <- get
  put $ st { constraints = NewFlex i (\ k' -> k' < k)
          -- DO NOT ADD constraints of form <= infty !!
          --               : Arc (Flex i) 0 (Rigid (RConst Infinite))
                         : constraints st }
 
addLeq :: Val -> Val -> TypeCheck ()
addLeq v1 v2 = -- trace ("Constraint: " ++ show v1 ++ " <= " ++ show v2) $
  do mc <- mkConstraint v1 v2 
     case mc of
       Nothing -> return ()
       Just c -> do
         st <- get
         put $ st { constraints = c : constraints st }
  
solveConstraints :: TypeCheck Solution
solveConstraints = do
  cs <- gets constraints
  case solve cs of 
    Just subst -> -- trace ("solution" ++ show subst) $ 
                    return subst 
    Nothing    -> fail $ "size constraints " ++ show cs ++ " unsolvable" 

-- solve constraints and substitute solution into the analyzed expression
solveAndModify :: Expr -> Env -> TypeCheck Expr
solveAndModify e rho = do
      sol <- solveConstraints
      let e' = subst (solToSubst sol rho) e
      resetConstraints
      return e'

nameOf :: Env -> Int -> Maybe Name
nameOf [] j = Nothing
nameOf ((x,VGen i):rho) j | i == j = Just x
nameOf (_:rho) j = nameOf rho j

sizeExprToExpr :: Env -> SizeExpr -> Expr
sizeExprToExpr rho (SizeConst Infinite) = Infty
sizeExprToExpr rho (SizeVar i n) | Just x <- nameOf rho i = add (Var x) n
  where add e n | n <= 0 = e
                | otherwise = add (Succ e) (n-1)

maxExpr :: [Expr] -> Expr
maxExpr [] = Infty
maxExpr [e] = e
maxExpr l = if Infty `elem` l then Infty else Max l

solToSubst :: Solution -> Env -> Subst
solToSubst sol rho = Map.map (maxExpr . map (sizeExprToExpr rho)) sol


{-
solToSubst :: Solution -> Env -> Subst
solToSubst sol rho = Map.foldWithKey step Map.empty sol
  where step k (SizeVar i n) sub | Just x <- nameOf rho i =
           Map.insert k (add (Var x) n) sub
        step k (SizeConst Infinite) sub = Map.insert k Infty sub
        step _ _ sub = sub

        add e n | n <= 0 = e
                | otherwise = add (Succ e) (n-1)
-}

-- pattern to Value ----------------------------------------------

{- RETIRED 
patternToVal :: Pattern -> TypeCheck Val
patternToVal p = do
  k <- getLen
  return $ fst (p2v k p)

-- turn a pattern into a value
-- dot patterns get variables corresponding to their flexible generic value
p2v :: Int -> Pattern -> (Val,Int)
p2v k p = 
    case p of
      VarP n -> (VGen k,k+1)
      ConP co n [] -> (VCon co n,k)
      ConP co n pl -> let (vl,k') = ps2vs k pl
                      in (VApp (VCon co n) vl,k')
      SuccP p -> let (v,k') = p2v k p
                 in (VSucc v,k')   
      DotP e -> (VGen k,k+1)

ps2vs :: Int -> [Pattern] -> ([Val],Int)
ps2vs k []  = ([],k)
ps2vs k (p:pl) = let (v,k') = p2v k p
                     (vl,k'') = ps2vs k' pl
                 in
                   (v:vl,k'')
-}